"use strict";(self.webpackChunkAutoGen_Studio=self.webpackChunkAutoGen_Studio||[]).push([[453],{70731:function(e,t,n){n.r(t);var o=n(96540),l=n(28007);const a={color:"#232129",padding:"96px",fontFamily:"-apple-system, Roboto, sans-serif, serif"},r={marginTop:0,marginBottom:64,maxWidth:320},i={marginBottom:48};t.default=()=>o.createElement("main",{style:a},o.createElement("title",null,"Not found"),o.createElement("h1",{style:r},"Page not found"),o.createElement("p",{style:i},"Sorry"," ",o.createElement("span",{role:"img","aria-label":"Pensive emoji"},"😔")," ","we couldn’t find what you were looking for.",o.createElement("br",null),null,o.createElement("br",null),o.createElement(l.Link,{to:"/"},"Go home"),"."))}}]);
//# sourceMappingURL=component---src-pages-404-tsx-9cfc00764f659a099247.js.map